/* bdb-gs.cp.c */

/* This program is the host processor software for a parallel version
of the Shur complement Gauss-Seidel iterative linear solver algorithm for
block-diagonal-bordered form sparse matrices.  */

/* special considerations for this implementation: 

1) Matrices are in block-diagonal-bordered form, but do not have to be
   symmetric.

2) any number of processors can be specified up to an including the
total number of processors on the CM-5 

3) it is assumed that ordering and load balancing are performed a
priori

*/
/*---------------------------------------------------------*/
#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <malloc.h>
#include <math.h>
#include <cm/cmmd.h>
#include "bdb_struct_par_gs.h"
#include "bdb_gs_MSG_NUM.h"
/*---------------------------------------------------------*/
#define YES        1
#define NO         0
/*---------------------------------------------------------*/
Proc           *proc;        /* proc[]       */
/*---------------------------------------------------------*/
Block          *blocks;      /* blocks[]      */
Block          *pblocks;     /* pblocks[]     */
/*---------------------------------------------------------*/
Diag           *diag;        /* diag[]        */
Diag           *pdiag;       /* pdiag[]       */
/*---------------------------------------------------------*/
BlockRow       *blockrow;    /* blockrow[]    */
BlockRow       *pblockrow;   /* pblockrow[]   */
/*---------------------------------------------------------*/
LowerBorder    *lborder;     /* lborder[]     */
LowerBorder    *plborder;    /* plborder[]    */
/*---------------------------------------------------------*/
LBorderRow     *lborderrow;  /* lborderrow[]  */
LBorderRow     *plborderrow; /* plborderrow[] */
/*---------------------------------------------------------*/
double         *b;           /* b[]         */
double         *pb;          /* pb[]        */
/*---------------------------------------------------------*/
int            n;            /* dimension of block-diagonal partition */
int            nblocks;      /* # of blocks             */
int            nblockrow;    /* dimension of blockrow   */
int            nlborder;     /* dimension of lborder    */
int            nlborderrow;  /* dimension of lborderrow */
/*---------------------------------------------------------*/
int            nprocs;
int            niter;
int            bytesreceived;
int            bytessent;
/*---------------------------------------------------------*/
LastBlockColor  **lbcolor;     /* lbcolor[][]  */
LastBlockColor   *lbcolortmp;  /* lbcolortmp[] */
LastBlockColor   *plbcolor;    /* plbcolor[]   */
/*---------------------------------------------------------*/
LastDiag         *ldiag;       /* ldiag[]      */
LastDiag         *pldiag;      /* pldiag[]     */
/*---------------------------------------------------------*/
LastBlockRow     *lblockrow;   /* lblockrow[]  */
LastBlockRow     *plblockrow;  /* plblockrow[] */
/*---------------------------------------------------------*/
double           *lb;          /* lb[]         */
int               ln;          /* dimension of the last block */
int               ncolors;
int               nlblockrow;
/*---------------------------------------------------------*/
UpdateLastBlock **updatelb; /* updatelb[][] */
UpdateLastBlock  *lbrbuf;   /* lbrbuf[] */
CommLB          **commlb;
CommLB           *commlbtmp;
int              *nupdates;
/*---------------------------------------------------------*/
/* function prototypes */
void   InputData();
int    min();

main(argc,argv)
  int    argc;
  char **argv;
{
  int i,j,l,bufsize;
  double ttimeCMMD,ttimeCMAML;

  printf("\n Executing host program... (pid=%d)\n", getpid());

  if (argc < 4)
   {
    printf("Usage: (1) input_file_name_A_matrix \n");
    printf("       (2) input_file_name_b_vector \n");
    printf("       (3) number of iterations before check convergence \n");
    exit(0);
   } /* end if */

  CMMD_enable();

/* input the data from the file - allocate memory as required */
  printf("input file for the A matrix  = |%s|\n",argv[1]);
  printf("input file for b vector  = |%s|\n",argv[2]);
  sscanf(argv[3],"%d",&niter);
  printf("number of iterations = |%d| \n",niter);
  fflush(stdout);

  printf("InputData() \n");
  fflush(stdout);
  InputData(argv[1],argv[2]);
  printf("End - InputData() --- nprocs=%d \n",nprocs);

if (nprocs == 1)
  {
    printf("Gauss-Seidel - 1 processor \n");
    fflush(stdout);
    CMMD_sync_host_with_nodes();
    CMMD_receive(CMMD_ANY_NODE,RECEIVE_TIME_GSS1,&node_tim,sizeof(node_timer));
    printf("Timing for pn=|%d| \telapsed=|%lf| \tbusy=|%lf| \tidle=|%lf| \n",
            node_tim.pn_id,node_tim.elapsed,node_tim.busy,node_tim.idle);

    printf("Gauss-Seidel - 1 processor - check convergence \n");
    fflush(stdout);
    CMMD_sync_host_with_nodes();
    CMMD_receive(CMMD_ANY_NODE,RECEIVE_TIME_GSC1,&node_tim,sizeof(node_timer));
    printf("Timing for pn=|%d| \telapsed=|%lf| \tbusy=|%lf| \tidle=|%lf| \n",
            node_tim.pn_id,node_tim.elapsed,node_tim.busy,node_tim.idle);

    printf("Gauss-Seidel - 1 processor - solve with %d iterations and convergence check \n",niter);
    fflush(stdout);
    CMMD_sync_host_with_nodes();
    CMMD_receive(CMMD_ANY_NODE,RECEIVE_TIME_GST1,&node_tim,sizeof(node_timer));
    printf("Timing for pn=|%d| \telapsed=|%lf| \tbusy=|%lf| \tidle=|%lf| \n",
            node_tim.pn_id,node_tim.elapsed,node_tim.busy,node_tim.idle);
    CMMD_sync_host_with_nodes();
  } /* end if nprocs == 1 */
else
  {
    printf("\nGauss-Seidel - diagonal blocks \n");
    fflush(stdout);
    CMMD_sync_host_with_nodes();
    for (i=0; i<nprocs; i++)
     {
      CMMD_receive(CMMD_ANY_NODE,RECEIVE_TIME_GSDB,&node_tim,sizeof(node_timer));
      printf("Timing for pn=|%d| \telapsed=|%lf| \tbusy=|%lf| \tidle=|%lf| \n",
              node_tim.pn_id,node_tim.elapsed,node_tim.busy,node_tim.idle);
      if (i == (nprocs-1))
	{
         ttimeCMMD = ttimeCMAML = node_tim.elapsed;
	} /* end if */
     } /* end for i */

    printf("\nGauss-Seidel - update last block - CMMD message passing \n");
    fflush(stdout);
    CMMD_sync_host_with_nodes();
    for (i=0; i<nprocs; i++)
     {
      CMMD_receive(CMMD_ANY_NODE,RECEIVE_TIME_GSULB,&node_tim,sizeof(node_timer));
      printf("Timing for pn=|%d| \telapsed=|%lf| \tbusy=|%lf| \tidle=|%lf| \n",
              node_tim.pn_id,node_tim.elapsed,node_tim.busy,node_tim.idle);
      if (i == (nprocs-1))
	{
         ttimeCMMD += node_tim.elapsed;
	} /* end if */
     } /* end for i */

    printf("Gauss-Seidel - last block - CMMD message passing \n");
    fflush(stdout);
    CMMD_sync_host_with_nodes();
    for (i=0; i<nprocs; i++)
     {
      CMMD_receive(CMMD_ANY_NODE,RECEIVE_TIME_GSLB,&node_tim,sizeof(node_timer));
      printf("Timing for pn=|%d| \telapsed=|%lf| \tbusy=|%lf| \tidle=|%lf| \n",
              node_tim.pn_id,node_tim.elapsed,node_tim.busy,node_tim.idle);
      if (i == (nprocs-1))
	{
         ttimeCMMD += node_tim.elapsed;
	} /* end if */
     } /* end for i */

    printf("Gauss-Seidel - check convergence - CMMD message passing \n");
    fflush(stdout);
    CMMD_sync_host_with_nodes();
    for (i=0; i<nprocs; i++)
     {
      CMMD_receive(CMMD_ANY_NODE,RECEIVE_TIME_GSCC,&node_tim,sizeof(node_timer));
      printf("Timing for pn=|%d| \telapsed=|%lf| \tbusy=|%lf| \tidle=|%lf| \n",
              node_tim.pn_id,node_tim.elapsed,node_tim.busy,node_tim.idle);
     } /* end for i */

    printf("\nGauss-Seidel - update last block - CMAML message passing \n");
    fflush(stdout);
    CMMD_sync_host_with_nodes();
    for (i=0; i<nprocs; i++)
     {
      CMMD_receive(CMMD_ANY_NODE,RECEIVE_TIME_GSULBAML,&node_tim,sizeof(node_timer));
      printf("Timing for pn=|%d| \telapsed=|%lf| \tbusy=|%lf| \tidle=|%lf| \n",
              node_tim.pn_id,node_tim.elapsed,node_tim.busy,node_tim.idle);
      if (i == (nprocs-1))
	{
         ttimeCMAML += node_tim.elapsed;
	} /* end if */
     } /* end for i */

    printf("Gauss-Seidel - last block - CMAML message passing \n");
    fflush(stdout);
    CMMD_sync_host_with_nodes();
    for (i=0; i<nprocs; i++)
     {
      CMMD_receive(CMMD_ANY_NODE,RECEIVE_TIME_GSLBAML,&node_tim,sizeof(node_timer));
      printf("Timing for pn=|%d| \telapsed=|%lf| \tbusy=|%lf| \tidle=|%lf| \n",
              node_tim.pn_id,node_tim.elapsed,node_tim.busy,node_tim.idle);
      if (i == (nprocs-1))
	{
         ttimeCMAML += node_tim.elapsed;
	} /* end if */
     } /* end for i */

    printf("Gauss-Seidel - check convergence - CMAML message passing \n");
    fflush(stdout);
    CMMD_sync_host_with_nodes();
    for (i=0; i<nprocs; i++)
     {
      CMMD_receive(CMMD_ANY_NODE,RECEIVE_TIME_GSCCAML,&node_tim,sizeof(node_timer));
      printf("Timing for pn=|%d| \telapsed=|%lf| \tbusy=|%lf| \tidle=|%lf| \n",
              node_tim.pn_id,node_tim.elapsed,node_tim.busy,node_tim.idle);
     } /* end for i */

    printf("\n\tTotal time for CMMD  message passing=|%lf| \n\tTotal time for CMAML message passing=|%lf|  \n",
              ttimeCMMD,ttimeCMAML);

    printf("\nGauss-Seidel - solve with %d iterations and one convergence check - CMMD message passing \n",niter);
    fflush(stdout);
    CMMD_sync_host_with_nodes();
    for (i=0; i<nprocs; i++)
     {
      CMMD_receive(CMMD_ANY_NODE,RECEIVE_TIME_GST,&node_tim,sizeof(node_timer));
      printf("Timing for pn=|%d| \telapsed=|%lf| \tbusy=|%lf| \tidle=|%lf| \n",
              node_tim.pn_id,node_tim.elapsed,node_tim.busy,node_tim.idle);
     } /* end for i */

    printf("\n\nGauss-Seidel - solve with %d iterations and one convergence check - CMAML message passing \n",niter);
    fflush(stdout);
    CMMD_sync_host_with_nodes();
    for (i=0; i<nprocs; i++)
     {
      CMMD_receive(CMMD_ANY_NODE,RECEIVE_TIME_GSTAML,&node_tim,sizeof(node_timer));
      printf("Timing for pn=|%d| \telapsed=|%lf| \tbusy=|%lf| \tidle=|%lf| \n",
              node_tim.pn_id,node_tim.elapsed,node_tim.busy,node_tim.idle);
     } /* end for i */

  } /* end if nprocs > 1 */

  printf("\nNumber of blocks = |%d| \n",(nblocks+1));
  printf("Total number of nonzeros = |%d| \n",
           (n+nblockrow+nlborderrow+ln+nlblockrow));
  printf("Fraction of nonzeros in the matrix = |%lf| \n\n",
          ((double)(n+nblockrow+nlborderrow+ln+nlblockrow) / (double)((n+ln)*(n+ln))) );
  fflush(stdout);

} /* end main() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void InputData(fnameA,fnameb)
  char *fnameA,*fnameb;
{
  int c,i,j,k,l,m,iproc,jproc,bufsize,icolor;
  int npblocks;     /* # of blocks distributed to a processor */
  int np;
  int npblockrow;   /* dimension of blockrow for a processor */
  int nplborder;    /* dimension of lborder for a processor */
  int nplborderrow; /* dimension of lborderrow for a processor */
  int npcolors;
  int npldiag;
  int nplblockrow;
  int lastblockstart,lastblockend;

  char *EOFCheck;
  FILE *fd, *fopen();

  EOFCheck = (char *) calloc(80,sizeof(char));
  if (EOFCheck == NULL)
   {
    printf("can't allocate memory for EOFCheck \n");
    exit(1);
   } /* end if */

  fd = fopen(fnameA,"r");
  if (fd == NULL)
   {
    printf("can't open file %s\n",fnameA);
    exit(1);
   } /* end if */

/* (I-0) read in the processor identifier array of structures */
  fscanf(fd,"%d",&nprocs);
printf("nprocs=|%d| \n",nprocs);
fflush(stdout);
  proc = (Proc *) calloc((nprocs+1),sizeof(Proc));
  if (proc == NULL)
   {
    printf("can't allocate memory for proc[] \n");
    exit(1);
   } /* end if */
  for (i=1; i<=nprocs; i++)
   {
    fscanf(fd,"%d %d %d %d %d %d %d ",&c,&c,
              &(proc[i].BlockPtr),
              &(proc[i].DiagPtr),
              &(proc[i].BlockRowPtr),
              &(proc[i].LowerBorderPtr),
              &(proc[i].LowerBorderRowPtr));
   } /* end for i */

/* (I-1) read in the block identifier array of structures */
  fscanf(fd,"%d",&nblocks);
printf("nblocks=|%d| \n",nblocks);
fflush(stdout);
  blocks = (Block *) calloc((nblocks+1),sizeof(Block));
  if (blocks == NULL)
   {
    printf("can't allocate memory for blocks[] \n");
    exit(1);
   } /* end if */
  for (i=1; i<=nblocks; i++)
   {
    fscanf(fd,"%d %d %d %d %d ",&c,
              &(blocks[i].BlockID),
              &(blocks[i].BlockSz),
              &(blocks[i].ProcNum),
              &(blocks[i].DiagPtr) );
   } /* end for i */

/* (I-2) read in the matrix diagonal array of structures */
  fscanf(fd,"%d",&n);
printf("n=|%d| \n",n);
fflush(stdout);
  diag = (Diag *) calloc((n+1),sizeof(Diag));
  if (diag == NULL)
   {
    printf("can't allocate memory for diag[] \n");
    exit(1);
   } /* end if */
  for (i=1; i<=n; i++)
   {
    fscanf(fd,"%d %lf %d ",&c,
              &(diag[i].Value),
              &(diag[i].PtrAix));
   } /* end for i */

/* (I-3) read in the block row storage array of structures */
  fscanf(fd,"%d",&nblockrow);
printf("nblockrow=|%d| \n",nblockrow);
fflush(stdout);
  blockrow = (BlockRow *) calloc((nblockrow+1),sizeof(BlockRow));
  if (blockrow == NULL)
   {
    printf("can't allocate memory for blockrow[] \n");
    exit(1);
   } /* end if */
  for (i=1; i<=nblockrow; i++)
   {
    fscanf(fd,"%d %lf %d %d ",&c,
              &(blockrow[i].Value),
              &(blockrow[i].i),
              &(blockrow[i].j));
   } /* end for i */

/* (I-4) read in the lower border row storage array of structures */
  fscanf(fd,"%d",&nlborder);
printf("nlborder=|%d| \n",nlborder);
fflush(stdout);
  lborder = (LowerBorder *) calloc((nlborder+1),sizeof(LowerBorder));
  if (lborder == NULL)
   {
    printf("can't allocate memory for lborder[] \n");
    exit(1);
   } /* end if */
  for (i=1; i<=nlborder; i++)
   {
    fscanf(fd,"%d %d %d %d %d ",&c,
              &(lborder[i].i),
              &(lborder[i].ulbproc),
              &(lborder[i].ulbloc),
              &(lborder[i].PtrLBix) );
   } /* end for i */

/* (I-5) read in the lower border row storage array of structures */
  fscanf(fd,"%d",&nlborderrow);
printf("nlborderrow=|%d| \n",nlborderrow);
fflush(stdout);
  lborderrow = (LBorderRow *) calloc((nlborderrow+1),sizeof(LBorderRow));
  if (lborderrow == NULL)
   {
    printf("can't allocate memory for lborderrow[] \n");
    exit(1);
   } /* end if */
  for (i=1; i<=nlborderrow; i++)
   {
    fscanf(fd,"%d %lf %d %d ",&c,
              &(lborderrow[i].Value),
              &(lborderrow[i].i),
              &(lborderrow[i].j));
   } /* end for i */

/* (I-7) read in the color identifier array of structures */
  fscanf(fd,"%d",&ncolors);
printf("ncolors=|%d| \n",ncolors);
fflush(stdout);
  lbcolortmp = (LastBlockColor *) calloc(((nprocs+1)*(ncolors+1)),sizeof(LastBlockColor));
  if (lbcolortmp == NULL)
   {
    printf("can't allocate memory for lbcolortmp[][] \n");
    exit(1);
   } /* end if */
  if (lbcolor != NULL) free(lbcolor);
  lbcolor = (LastBlockColor **) calloc((ncolors+1),sizeof(LastBlockColor *));
  if (lbcolor == NULL)
   {
    printf("can't allocate memory for lbcolor[][] \n");
    exit(1);
   } /* end if */
  for (icolor=0;icolor<=ncolors;icolor++) lbcolor[icolor] = lbcolortmp + (icolor*(nprocs+1));

/* the graph coloring routine will break the last block into processors 
  and colors within processors */
 for (icolor=1; icolor<=ncolors; icolor++)
    for (iproc=1;iproc<=nprocs;iproc++)
     {
      fscanf(fd,"%d %d %d %d %d ",&c,
                &(lbcolor[icolor][iproc].Color),
                &(lbcolor[icolor][iproc].ProcNum),
                &(lbcolor[icolor][iproc].Starti),
                &(lbcolor[icolor][iproc].DiagPtr));
     } /* end for iproc & icolor */

/* (I-8) read in the last block diagonal array of structures */
  fscanf(fd,"%d",&ln);
printf("ln=|%d| \n",ln);
fflush(stdout);
  ldiag = (LastDiag *) calloc((ln+1),sizeof(LastDiag));
  if (ldiag == NULL)
   {
    printf("can't allocate memory for ldiag[] \n");
    exit(1);
   } /* end if */
  for (i=1; i<=ln; i++)
   {
    fscanf(fd,"%d %lf %d ",&c,
              &(ldiag[i].Value),
              &(ldiag[i].PtrAix));
   } /* end for i */

/* (I-9) read in the last block row storage array of structures */
  fscanf(fd,"%d",&nlblockrow);
printf("nlblockrow=|%d| \n",nlblockrow);
fflush(stdout);
  lblockrow = (LastBlockRow *) calloc((nlblockrow+1),sizeof(LastBlockRow));
  if (lblockrow == NULL)
   {
    printf("can't allocate memory for lblockrow[] \n");
    exit(1);
   } /* end if */
  for (i=1; i<=nlblockrow; i++)
   {
    fscanf(fd,"%d %lf %d %d ",&c,
              &(lblockrow[i].Value),
              &(lblockrow[i].i),
              &(lblockrow[i].j));
  } /* end for i */

/* (II-0) load the block structure buffers for the independent blocks */
  lastblockstart = n + 1;
  lastblockend   = n + ln;

printf("lastblockend=%d - lastblockstart=%d \n",lastblockend,lastblockstart);
fflush(stdout);
  bbuf.nprocs         = nprocs;
  bbuf.n              = n;
  bbuf.nblocks        = nblocks;
  bbuf.nblockrow      = nblockrow;
  bbuf.nlborder       = nlborder;
  bbuf.nlborderrow    = nlborderrow;
  bbuf.ncolors        = ncolors;
  bbuf.ln             = ln;
  bbuf.nlblockrow     = nlblockrow;
  bbuf.niter          = niter;
  bbuf.lastblockstart = lastblockstart;
  bbuf.lastblockend   = lastblockend;
  bufsize = (12*sizeof(int));
  CMMD_bc_from_host(&bbuf,bufsize);

/* (II-1) load the block structure buffers for the independent blocks */
  for (i=0; i<nprocs; i++)
   {
    l = proc[i].BlockPtr;
    m = proc[i+1].BlockPtr;
    npblocks = (m-l);

printf("load the block structure buffers for the independent blocks npblocks=%d - l=%d - m=%d \n",
              npblocks,l,m);
fflush(stdout);

    CMMD_send(i,SEND_NPBLOCKS,&npblocks,sizeof(int));
    pblocks = (Block *) calloc((npblocks+1),sizeof(Block));
    if (pblocks == NULL)
     {
      printf("can't allocate memory for pblocks[] - proc - %d \n",i);
      exit(1);
     } /* end if */
    pblocks[0].BlockID = 0;
    pblocks[0].BlockSz = 0;
    pblocks[0].ProcNum = -1;
    pblocks[0].DiagPtr = proc[i].DiagPtr;

printf("\ti=%d - l=%d - m=%d - pblocks[0].DiagPtr=%d \n",i,l,m,pblocks[0].DiagPtr);
fflush(stdout);
/*
if (i==1) {
  printf("pblocks[0].BlockID=%d - \t.BlockSz=%d - \t.ProcNum=%d - \t.DiagPtr=%d \n",
            pblocks[0].BlockID,
            pblocks[0].BlockSz,
            pblocks[0].ProcNum,
            pblocks[0].DiagPtr);
  fflush(stdout); }
*/
    for (j=1; j<=npblocks; j++)
     {
      pblocks[j].BlockID = blocks[j+l].BlockID;
      pblocks[j].BlockSz = blocks[j+l].BlockSz;
      pblocks[j].ProcNum = blocks[j+l].ProcNum;
      pblocks[j].DiagPtr = blocks[j+l].DiagPtr;
/*
if (i==1) {
  printf("pblocks[%d].BlockID=%d - \t.BlockSz=%d - \t.ProcNum=%d - \t.DiagPtr=%d \n",
          j,pblocks[j].BlockID,
            pblocks[j].BlockSz,
            pblocks[j].ProcNum,
            pblocks[j].DiagPtr);
  fflush(stdout); }
*/
     } /* end for j */
/*
printf("send the block structure buffers for the independent blocks \n");
fflush(stdout);
*/
    bufsize = ((npblocks+1)*sizeof(Block));
    CMMD_send(i,SEND_PBLOCKS,pblocks,bufsize);
    free(pblocks); 
   } /* end for i */

/* (II-2) load the matrix diagonal structure buffers for the independent blocks */
printf("load the matrix diagonal structure buffers for the independent blocks \n");
fflush(stdout);
  for (i=0; i<nprocs; i++)
   {
    l = proc[i].DiagPtr;
    m = proc[i+1].DiagPtr;
    np = (m-l);
    CMMD_send(i,SEND_NP,&np,sizeof(int));
    pdiag = (Diag *) calloc((np+1),sizeof(Diag));
    if (pdiag == NULL)
     {
      printf("can't allocate memory for pdiag[] - proc - %d \n",i);
      exit(1);
     } /* end if */
    pdiag[0].Value    = 0.0;
    pdiag[0].PtrAix   = 0;
/*
if (i==1) {
  printf("pdiag[0].Value=%lf - \t.PtrAix=%d \n",
            pdiag[0].Value,
            pdiag[0].PtrAix);
  fflush(stdout); }
*/
    for (j=1; j<=np; j++)
     {
      pdiag[j].Value    = diag[j+l].Value;
      pdiag[j].PtrAix   = diag[j+l].PtrAix - proc[i].BlockRowPtr;
/*
if (i==1) {
  printf("pdiag[%d].Value=%lf - \t.PtrAix=%d \n",
          j,pdiag[j].Value,
            pdiag[j].PtrAix);
  fflush(stdout); }
*/
     } /* end for j */
/*
printf("send the matrix diagonal structure buffers for the independent blocks \n");
fflush(stdout);
*/
    bufsize = ((np+1)*sizeof(Diag));
    CMMD_send(i,SEND_PDIAG,pdiag,bufsize);
    free(pdiag);
   } /* end for i */

/* (II-3) load the block row storage structure buffers for the independent blocks */
printf("load the block row storage structure buffers for the independent blocks \n");
fflush(stdout);
  for (i=0; i<nprocs; i++)
   {
    l = proc[i].BlockRowPtr;
    m = proc[i+1].BlockRowPtr;
    npblockrow = (m-l);
    CMMD_send(i,SEND_NPBLOCKROW,&npblockrow,sizeof(int));
    pblockrow = (BlockRow *) calloc((npblockrow+1),sizeof(BlockRow));
    if (pblockrow == NULL)
     {
      printf("can't allocate memory for pblockrow[] - proc - %d \n",i);
      exit(1);
     } /* end if */
    pblockrow[0].Value   = 0.0;
    pblockrow[0].i       = 0;
    pblockrow[0].j       = 0;
/*
if (i==1) {
  printf("pblockrow[0].Value=%lf - \t.i=%d - \t.j=%d \n",
          pblockrow[0].Value,
          pblockrow[0].i,
          pblockrow[0].j);
  fflush(stdout); }
*/
    for (j=1; j<=npblockrow; j++)
     {
      pblockrow[j].Value   = blockrow[j+l].Value;
      pblockrow[j].i       = blockrow[j+l].i;
      pblockrow[j].j       = blockrow[j+l].j;
/*
if (i==1) {
  printf("pblockrow[%d].Value=%lf - \t.i=%d - \t.j=%d \n",
        j,pblockrow[j].Value,
          pblockrow[j].i,
          pblockrow[j].j);
  fflush(stdout); }
*/
     } /* end for j */

printf("send the block row storage structure buffers for the independent blocks --- iproc=%d - npblockrow=%d -\n",
          i,npblockrow);
fflush(stdout);

    bufsize = ((npblockrow+1)*sizeof(BlockRow));
    CMMD_send(i,SEND_PBLOCKROW,pblockrow,bufsize);
    free(pblockrow);
   } /* end for i */

/* (II-4) load the lower border storage buffers for the independent blocks */
printf("load the lower border storage buffers for the independent blocks \n");
fflush(stdout);
  for (i=0; i<nprocs; i++)
   {
    l = proc[i].LowerBorderPtr;
    m = proc[i+1].LowerBorderPtr;
    nplborder = (m-l);
    CMMD_send(i,SEND_NPLBORDER,&nplborder,sizeof(int));
    plborder = (LowerBorder *) calloc((nplborder+1),sizeof(LowerBorder));
    if (plborder == NULL)
     {
      printf("can't allocate memory for plborder[] - proc - %d \n",i);
      exit(1);
     } /* end if */
    plborder[0].i       = 0;
    plborder[0].ulbproc = 0;
    plborder[0].ulbloc  = 0;
    plborder[0].PtrLBix = 0;
/*
if (i==1) {
  printf("plborder[0].i=%d - \t.ulbproc=%d - \t.ulbloc=%d - \t.PtrLBix=%d \n",
          plborder[0].i,
          plborder[0].ulbproc,
          plborder[0].ulbloc,
          plborder[0].PtrLBix);
  fflush(stdout); }
*/
    for (j=1; j<=nplborder; j++)
     {
      plborder[j].i       = lborder[j+l].i;
      plborder[j].ulbproc = lborder[j+l].ulbproc - 1;
      plborder[j].ulbloc  = lborder[j+l].ulbloc - 1;
      plborder[j].PtrLBix = lborder[j+l].PtrLBix - proc[i].LowerBorderRowPtr;
/*
if (i==1) {
  printf("plborder[%d].i=%d - \t.ulbproc=%d - \t.ulbloc=%d - \t.PtrLBix=%d \n",
        j,plborder[j].i,
          plborder[j].ulbproc,
          plborder[j].ulbloc,
          plborder[j].PtrLBix);
  fflush(stdout); }
*/
     } /* end for j */
/*
printf("send the lower border storage buffers for the independent blocks \n");
fflush(stdout);
*/
    bufsize = ((nplborder+1)*sizeof(LowerBorder));
    CMMD_send(i,SEND_PLBORDER,plborder,bufsize);
    free(plborder);
   } /* end for i */

/* (II-5) load the lower border row structure buffers for the independent blocks */
printf("load the lower border row structure buffers for the independent blocks \n");
fflush(stdout);
  for (i=0; i<nprocs; i++)
   {
    l = proc[i].LowerBorderRowPtr;
    m = proc[i+1].LowerBorderRowPtr;
    nplborderrow = (m-l);
    CMMD_send(i,SEND_NPLBORDERROW,&nplborderrow,sizeof(int));
    plborderrow = (LBorderRow *) calloc((nplborderrow+1),sizeof(LBorderRow));
    if (plborderrow == NULL)
     {
      printf("can't allocate memory for plborderrow[] - proc - %d \n",i);
      exit(1);
     } /* end if */
    plborderrow[0].Value    = 0.0;
    plborderrow[0].i        = 0;
    plborderrow[0].j        = 0;
/*
if (i==1) {
  printf("plborderrow[0].Value=%lf - \t.i=%d - \t.j=%d \n",
             plborderrow[0].Value,
             plborderrow[0].i,
             plborderrow[0].j);
  fflush(stdout); }
*/
    for (j=1; j<=nplborderrow; j++)
     {
      plborderrow[j].Value    = lborderrow[j+l].Value;
      plborderrow[j].i        = lborderrow[j+l].i;
      plborderrow[j].j        = lborderrow[j+l].j;
/*
if (i==1) {
  printf("plborderrow[%d].Value=%lf - \t.i=%d - \t.j=%d \n",
           j,plborderrow[j].Value,
             plborderrow[j].i,
             plborderrow[j].j);
  fflush(stdout); }
*/
     } /* end for j */
/*
printf("send the lower border row structure buffers for the independent blocks \n");
fflush(stdout);
*/
    bufsize = ((nplborderrow+1)*sizeof(LBorderRow));
    CMMD_send(i,SEND_PLBORDERROW,plborderrow,bufsize);
    free(plborderrow);
   } /* end for i */
  free(lborderrow);

/* (II-7) load the color identifier structure buffers for the independent blocks */
printf("load the color identifier structure buffers for the independent blocks \n");
fflush(stdout);
   
 for (icolor=1; icolor<=ncolors; icolor++)
    for (iproc=1; iproc<=nprocs; iproc++)
     {
      lbcolor[icolor][iproc].RowPtr = ldiag[ lbcolor[icolor][iproc].DiagPtr ].PtrAix;

  printf("lbcolor[%d][%d].RowPtr=%d - lbcolor[%d][%d].DiagPtr=%d - ldiag[%d].PtrAix=%d - lbcolor[%d][%d].Starti=%d \n",
        icolor,iproc,lbcolor[icolor][iproc].RowPtr,
        icolor,iproc,lbcolor[icolor][iproc].DiagPtr,
        lbcolor[icolor][iproc].DiagPtr,ldiag[ lbcolor[icolor][iproc].DiagPtr ].PtrAix,
        icolor,iproc,lbcolor[icolor][iproc].Starti);
  fflush(stdout);

     } /* end for iproc & icolor */
  
  lbcolor[0][0].Color   = 0;
  lbcolor[0][0].ProcNum = 0;
  lbcolor[0][0].Starti  = 0;
  lbcolor[0][0].DiagPtr = 0;
  lbcolor[0][0].RowPtr  = 0;
/*
  printf("lbcolor[0][0].Color=%d -\t.ProcNum=%d - \t.Starti=%d - \t.DiagPtr=%d - \t.RowPtr=%d \n",
                     lbcolor[0][0].Color,
                     lbcolor[0][0].ProcNum,
                     lbcolor[0][0].Starti,
                     lbcolor[0][0].DiagPtr,
                     lbcolor[0][0].RowPtr);
  fflush(stdout); 
*/
  for (icolor=1;icolor<=ncolors;icolor++)
   {
    lbcolor[icolor][0].Color   = icolor;
    lbcolor[icolor][0].ProcNum = 0;
    lbcolor[icolor][0].Starti  = 0;
    if (icolor == 1)
      {
       lbcolor[icolor][0].DiagPtr = 0;
       lbcolor[icolor][0].RowPtr  = 0;
       } /* end if */
    else
      {
       lbcolor[icolor][0].DiagPtr = lbcolor[icolor-1][nprocs].DiagPtr;
       lbcolor[icolor][0].RowPtr  = lbcolor[icolor-1][nprocs].RowPtr;
      } /* end else */
/*
printf("lbcolor[%d][0].Color=%d -\t.ProcNum=%d - \t.Starti=%d - \t.DiagPtr=%d - \t.RowPtr=%d \n",
        icolor,lbcolor[icolor][0].Color,
               lbcolor[icolor][0].ProcNum,
               lbcolor[icolor][0].Starti,
               lbcolor[icolor][0].DiagPtr,
               lbcolor[icolor][0].RowPtr);
fflush(stdout); 
*/
   } /* end for icolor  */

  for (iproc=1; iproc<=nprocs; iproc++)
   {
    plbcolor = (LastBlockColor *) calloc((ncolors+1),sizeof(LastBlockColor));
    if (plbcolor == NULL)
     {
      printf("can't allocate memory for plbcolor[] - proc - %d \n",i);
      exit(1);
     } /* end if */

    plbcolor[0].Color    = 0;
    plbcolor[0].ProcNum  = 0;
    plbcolor[0].Starti   = lbcolor[0][iproc].Starti;
    plbcolor[0].DiagPtr  = lbcolor[0][iproc].DiagPtr;
/*
printf("iproc=%d - plbcolor[0].Color=%d - \t.ProcNum=%d - \t.Starti=%d - \t.DiagPtr=%d \n",
         (iproc-1),plbcolor[0].Color,
               plbcolor[0].ProcNum,
               plbcolor[0].Starti,
               plbcolor[0].DiagPtr);
fflush(stdout);
*/
    for (icolor=1; icolor<=ncolors; icolor++)
     {
      plbcolor[icolor].Color   = lbcolor[icolor][iproc].Color;
      plbcolor[icolor].ProcNum = (lbcolor[icolor][iproc].ProcNum - 1);
      if (icolor == 1)
        plbcolor[icolor].Starti  = 1;
      else
        plbcolor[icolor].Starti  = plbcolor[icolor-1].Starti + 
                                    (lbcolor[icolor-1][iproc].DiagPtr - lbcolor[icolor-1][iproc-1].DiagPtr);
      if (icolor == 1)
        plbcolor[icolor].DiagPtr = (lbcolor[icolor][iproc].DiagPtr - lbcolor[icolor][iproc-1].DiagPtr);
      else
        plbcolor[icolor].DiagPtr = plbcolor[icolor-1].DiagPtr +
                                   (lbcolor[icolor][iproc].DiagPtr - lbcolor[icolor][iproc-1].DiagPtr);
/*
printf("iproc=%d - plbcolor[%d].Color=%d - \t.ProcNum=%d - \t.Starti=%d - \t.DiagPtr=%d \n",
        (iproc-1),icolor,plbcolor[icolor].Color,
                     plbcolor[icolor].ProcNum,
                     plbcolor[icolor].Starti,
                     plbcolor[icolor].DiagPtr);
fflush(stdout);
*/
     } /* end for icolor */
/*
printf("send the matrix diagonal structure buffers for the independent blocks \n");
fflush(stdout);
*/
    bufsize = ((npcolors+1)*sizeof(LastBlockColor));
    CMMD_send((iproc-1),SEND_PLBCOLORS,plbcolor,bufsize);
/*
printf("send the matrix diagonal structure buffers for the independent blocks \n");
fflush(stdout);
*/
    free(plbcolor);
   } /* end for iproc */

/* (II-8) load the last block diagonal structure buffers for the independent blocks */
printf("load the last block diagonal structure buffers for the independent blocks \n");
fflush(stdout);
  for (iproc=1; iproc<=nprocs; iproc++)
   {
/* calculate the number of rows for this proc */
    npldiag = 0;
    for (icolor=1;icolor<=ncolors;icolor++)
      {
       npldiag += lbcolor[icolor][iproc].DiagPtr - lbcolor[icolor][iproc-1].DiagPtr;
      } /* end for icolor */

printf("iproc=%d - npldiag=%d \n",iproc,npldiag);
fflush(stdout); 

    CMMD_send((iproc-1),SEND_NPLDIAG,&npldiag,sizeof(int));

    pldiag = (LastDiag *) calloc((npldiag+1),sizeof(LastDiag));
    if (pldiag == NULL)
     {
      printf("can't allocate memory for pldiag[] - proc - %d \n",i);
      exit(1);
     } /* end if */

    i = 0;
    pldiag[i].Value  = 0.0;
    pldiag[i].i      = 0;
    pldiag[i].PtrAix = 0;
/*
printf("\tpldiag[%d].Value=%lf - \t.i=%d - \t.PtrAix=%d \n",
          i,pldiag[i].Value,
            pldiag[i].i,
            pldiag[i].PtrAix);
fflush(stdout); 
*/
    for (icolor=1;icolor<=ncolors;icolor++)
      {
       l = (lbcolor[icolor][iproc-1].DiagPtr+1);
       m = lbcolor[icolor][iproc].DiagPtr;
       for (j=l; j<=m; j++)
        {
         i++;
         pldiag[i].Value  = ldiag[j].Value;
         pldiag[i].i      = (j + lastblockstart - 1);
         pldiag[i].PtrAix = pldiag[i-1].PtrAix + (ldiag[j].PtrAix - ldiag[j-1].PtrAix);
/*
printf("\tpldiag[%d].Value=%lf - \t.i=%d - \t.PtrAix=%d \n",
        i,pldiag[i].Value,
          pldiag[i].i,
          pldiag[i].PtrAix);
fflush(stdout); 
*/
        } /* end for j */
      } /* end for icolor */
/*
printf("send the last block diagonal structure buffers for the independent blocks \n");
fflush(stdout);
*/

    bufsize = ((npldiag+1)*sizeof(LastDiag));
    CMMD_send((iproc-1),SEND_PLDIAG,pldiag,bufsize);

    free(pldiag);
   } /* end for iproc */

/* (II-9) load the last block row storage structure buffers for the independent blocks */
printf("load the last block row storage structure buffers for the independent blocks \n");
fflush(stdout);
  for (iproc=1; iproc<=nprocs; iproc++)
   {
/* calculate the number of rows elements for this proc */
    nplblockrow = 0;
    for (icolor=1;icolor<=ncolors;icolor++)
      {
       nplblockrow += lbcolor[icolor][iproc].RowPtr - lbcolor[icolor][iproc-1].RowPtr;
      } /* end for icolor */
/*
printf("iproc=%d - nplblockrow=%d \n",iproc,nplblockrow);
fflush(stdout); 
*/
    CMMD_send((iproc-1),SEND_NPLBLOCKROW,&nplblockrow,sizeof(int));

    plblockrow = (LastBlockRow *) calloc((nplblockrow+1),sizeof(LastBlockRow));
    if (plblockrow == NULL)
     {
      printf("can't allocate memory for plblockrow[] - proc - %d \n",iproc);
      exit(1);
     } /* end if */

    i = 0;
    plblockrow[i].Value   = 0.0;
    plblockrow[i].i       = 0;
    plblockrow[i].j       = 0;
/*
printf("\tplblockrow[%d].Value=%lf - \t.i=%d - \t.j=%d \n",
          i,plblockrow[i].Value,
            plblockrow[i].i,
            plblockrow[i].j);
fflush(stdout); 
*/
    for (icolor=1;icolor<=ncolors;icolor++)
      {
       l = (lbcolor[icolor][iproc-1].RowPtr+1);
       m = lbcolor[icolor][iproc].RowPtr;
       for (j=l; j<=m; j++)
        {
         i++;
         plblockrow[i].Value   = lblockrow[j].Value;
         plblockrow[i].i       = lblockrow[j].i;
         plblockrow[i].j       = lblockrow[j].j;
/*
printf("\tplblockrow[%d].Value=%lf - \t.i=%d  - \t.j=%d\n",
        i,plblockrow[i].Value,
          plblockrow[i].i,
          plblockrow[i].j);
fflush(stdout); 
*/
        } /* end for j */
      } /* end for icolor */
/*
printf("send the last block row storage structure buffers for the independent blocks \n");
fflush(stdout);
*/

    bufsize = ((nplblockrow+1)*sizeof(LastBlockRow));
    CMMD_send((iproc-1),SEND_PLBLOCKROW,plblockrow,bufsize);

    free(plblockrow);
   } /* end for iproc */

/* (II-10) send the sizes of messages to update the last block for each processor */

printf("send the sizes of messages to update the last block for each processor \n");
fflush(stdout);

  ulbbufsize = (int **) calloc(nprocs,sizeof(int *));
  if (ulbbufsize == NULL)
   {
    printf("can't allocate memory for ulbbufsize[][] \n");
    exit(1);
   } /* end if */
  ulbbufsizetmp = (int *) calloc((nprocs*nprocs),sizeof(int));
  if (ulbbufsizetmp == NULL)
   {
    printf("can't allocate memory for ulbbufsizetmp[] \n");
    exit(1);
   } /* end if */
  for (iproc=0;iproc<nprocs;iproc++) ulbbufsize[iproc] = ulbbufsizetmp + (iproc * nprocs);

  for (iproc=0;iproc<nprocs;iproc++)
    {
     fscanf(fd,"%d %d ",&c,&c);
     for (jproc=0;jproc<nprocs;jproc++)
      {
       fscanf(fd,"%d %d %d ",&c,&c,&(ulbbufsize[iproc][jproc]));

printf("iproc=%d - jproc=%d - \tulbbufsize[%d][%d]=%d \n",iproc,jproc,iproc,jproc,ulbbufsize[iproc][jproc]);
fflush(stdout); 

      } /* end for jproc */
    } /* end for iproc */
 
  for (iproc=0;iproc<nprocs;iproc++)
    {
      bufsize = ((nprocs*nprocs)*sizeof(int));
      CMMD_send(iproc,SEND_PULBBUF,ulbbufsizetmp,bufsize);
    } /* end for iproc */

/* (II-11) read in a character string to identify EOF */
  fscanf(fd,"%s",EOFCheck);
  printf("Input_A_matrix - |%s| \n",EOFCheck);
  fflush(stdout);

  fclose(fd);

  fd = fopen(fnameb,"r");
  if (fd == NULL)
   {
    printf("can't open file %s\n",fnameb);
    exit(1);
   } /* end if */

/* (II-12) read in the b vector */
printf("read in the b vector --- n=%d \n",n);
fflush(stdout);
  b = (double *) calloc((n+1),sizeof(double));
  if (b == NULL)
   {
    printf("can't allocate memory for b[] \n");
    exit(1);
   } /* end if */
  for (i=1; i<=n; i++)
   {
    fscanf(fd,"%lf",&(b[i]));
   } /* end for i */
/* load the b vector buffers for the independent blocks */
printf("load the b vector buffers for the independent blocks \n");
fflush(stdout);
  for (iproc=0; iproc<nprocs; iproc++)
   {
    l = proc[iproc].DiagPtr;
    m = proc[iproc+1].DiagPtr;
    np = (m-l);
/*
printf("\tiproc=%d - l=%d - m=%d - np=%d \n",iproc,l,m,np);
fflush(stdout);
*/
    pb = (double *) calloc((np+1),sizeof(double));
    if (pb == NULL)
     {
      printf("can't allocate memory for pb[] - proc - %d \n",i);
      exit(1);
     } /* end if */
    for (j=1; j<=np; j++)
     {
      pb[j] = b[j+l];
     } /* end for j */
/*
printf("load the b vector buffers for the independent blocks \n");
fflush(stdout);
*/

    bufsize = (np*sizeof(double));
    CMMD_send(iproc,SEND_PB,&(pb[1]),bufsize);

    free(pb);
   } /* end for iproc */

/* (II-13) read in the lb vector */
printf("load the lb vector buffers for the independent blocks --- ln=%d \n",ln);
fflush(stdout);
  lb = (double *) calloc((ln+1),sizeof(double));
  if (lb == NULL)
   {
    printf("can't allocate memory for lb[] \n");
    exit(1);
   } /* end if */
  for (i=0; i<ln; i++)
   {
    fscanf(fd,"%lf",&(lb[i]));
   } /* end for i */

/* load the lb vector onto the processors */
printf("load the lb vector onto the processors \n");
fflush(stdout);
  for (iproc=0; iproc<nprocs; iproc++)
   {
/*
printf("send the lb vector onto the processors i=%d \n",i);
fflush(stdout);
*/

    bufsize = (ln*sizeof(double));
    CMMD_send(iproc,SEND_LB,lb,bufsize);

   } /* end for iproc */

/* (II-14) read in a character string to identify EOF */
  fscanf(fd,"%s",EOFCheck);
  printf("Input_b_vector - |%s| \n",EOFCheck);
  fflush(stdout);

  fclose(fd);

/* (II-16) send the last block communication identifiers */
printf("send the last block communication identifiers \n");
fflush(stdout);

  commlbtmp = (CommLB *) calloc(((nprocs+1)*(ncolors+1)),sizeof(CommLB));
  if (commlbtmp == NULL)
   {
    printf("can't allocate memory for commlbtmp[] \n");
    exit(1);
   } /* end if */
  commlb = (CommLB **) calloc((ncolors+1),sizeof(CommLB *));
  if (commlb == NULL)
   {
    printf("can't allocate memory for commlb[][] \n");
    exit(1);
   } /* end if */
  for (icolor=0;icolor<=ncolors;icolor++) commlb[icolor] = commlbtmp + icolor*(nprocs+1);

printf("load commlb[][] \n");
fflush(stdout);

  for (icolor=1;icolor<=ncolors;icolor++)
    {
     for (iproc=1;iproc<nprocs;iproc++)
        {
         commlb[icolor][iproc].starti  = lbcolor[icolor][iproc].Starti;
         commlb[icolor][iproc].bufsize = lbcolor[icolor][iproc+1].Starti - lbcolor[icolor][iproc].Starti;
        } /* end for icolor */
     commlb[icolor][nprocs].starti  = lbcolor[icolor][nprocs].Starti;
     if (icolor < ncolors)
        commlb[icolor][nprocs].bufsize = lbcolor[icolor+1][1].Starti - lbcolor[icolor][nprocs].Starti;
     else
       commlb[icolor][nprocs].bufsize = (ln - lbcolor[icolor][nprocs].Starti + 1);
    } /* end for iproc */
/*
for (icolor=1;icolor<=ncolors;icolor++)
  for (iproc=0;iproc<=nprocs;iproc++) {
  printf("\t\tcommlb[%d][%d].starti=%d - \t.bufsize=%d \n",
         icolor,iproc,commlb[icolor][iproc].starti,
                      commlb[icolor][iproc].bufsize);
  fflush(stdout); }
*/
  bufsize = ((nprocs+1)*(ncolors+1)*sizeof(CommLB));
  for (iproc=0;iproc<nprocs;iproc++)
    {
     CMMD_send(iproc,SEND_PULBBUF,commlbtmp,bufsize);
    } /* end for iproc */

  nupdates = (int *) calloc((nprocs+1),sizeof(int));
  if (nupdates == NULL)
   {
    printf("can't allocate memory for nupdates[] \n");
    exit(1);
   } /* end if */
  for (i=0;i<=nprocs;i++) nupdates[i] = 0;
  for (i=1;i<=nlborder;i++)
    {
/*
printf("\t\ti=%d - lborder[%d].ulbproc]=%d \n",i,i,lborder[i].ulbproc);
fflush(stdout);
*/
     nupdates[lborder[i].ulbproc]++;
    } /* end for i */
  for (i=0;i<=nprocs;i++) 
    printf("\t\tiproc=%d - nupdates[%d]=%d \n",i,i,nupdates[i]);
  fflush(stdout);

  bufsize = ((nprocs+1)*sizeof(int));
  for (iproc=0;iproc<nprocs;iproc++)
    {
     CMMD_send(iproc,SEND_NUPDATE,nupdates,bufsize);
    } /* end for iproc */

} /* end InputData() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
int min(i,j)
 int i,j;
{
  if (i<=j) return i;
   else return j;

} /* end min() */
